"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.Metaplex = void 0;
const web3_js_1 = require("@solana/web3.js");
const drivers_1 = require("./drivers");
const shared_1 = require("./shared");
const modules_1 = require("./modules");
class Metaplex {
    constructor(connection, options = {}) {
        /** The registered handlers for read/write operations. */
        this.operationHandlers = new Map();
        this.connection = connection;
        this.options = options;
        this.identityDriver = new drivers_1.GuestIdentityDriver(this);
        this.storageDriver = new drivers_1.BundlrStorageDriver(this);
        this.registerDefaultPlugins();
    }
    static make(connection, options = {}) {
        return new this(connection, options);
    }
    registerDefaultPlugins() {
        this.use((0, modules_1.nftPlugin)());
    }
    use(plugin) {
        plugin.install(this);
        return this;
    }
    identity() {
        return this.identityDriver;
    }
    setIdentity(identity) {
        this.identityDriver = identity;
        return this;
    }
    storage() {
        return this.storageDriver;
    }
    setStorage(storage) {
        this.storageDriver = storage;
        return this;
    }
    sendTransaction(transaction, signers = [], sendOptions = {}) {
        return __awaiter(this, void 0, void 0, function* () {
            if (transaction instanceof shared_1.TransactionBuilder) {
                signers = [...transaction.getSigners(), ...signers];
                transaction = transaction.toTransaction();
            }
            const { keypairs, identities } = (0, shared_1.getSignerHistogram)(signers);
            for (let i = 0; i < identities.length; i++) {
                if (!identities[i].is(this.identity())) {
                    yield identities[i].signTransaction(transaction);
                }
            }
            return this.identity().sendTransaction(transaction, keypairs, sendOptions);
        });
    }
    confirmTransaction(signature, commitment) {
        return __awaiter(this, void 0, void 0, function* () {
            const rpcResponse = yield this.connection.confirmTransaction(signature, commitment);
            let transaction_error = rpcResponse.value.err;
            if (transaction_error) {
                // TODO: Custom errors.
                throw new web3_js_1.SendTransactionError(`Transaction ${signature} failed (${JSON.stringify(transaction_error)})`, [transaction_error.toString()]);
            }
            return rpcResponse;
        });
    }
    sendAndConfirmTransaction(transaction, signers, confirmOptions) {
        return __awaiter(this, void 0, void 0, function* () {
            const signature = yield this.sendTransaction(transaction, signers, confirmOptions);
            yield this.confirmTransaction(signature, confirmOptions === null || confirmOptions === void 0 ? void 0 : confirmOptions.commitment);
            return signature;
        });
    }
    getAccountInfo(publicKey, commitment) {
        return __awaiter(this, void 0, void 0, function* () {
            return this.connection.getAccountInfo(publicKey, commitment);
        });
    }
    getMultipleAccountsInfo(publicKeys, commitment) {
        return __awaiter(this, void 0, void 0, function* () {
            const accounts = yield this.connection.getMultipleAccountsInfo(publicKeys, commitment);
            return accounts;
        });
    }
    register(operation, operationHandler) {
        this.operationHandlers.set(operation, operationHandler);
        return this;
    }
    getOperationHandler(operation) {
        const operationHandler = this.operationHandlers.get(operation.constructor);
        if (!operationHandler) {
            // TODO: Custom errors.
            throw new Error(`No operation handler registered for ${operation.constructor.name}`);
        }
        return new operationHandler(this);
    }
    execute(operation) {
        return __awaiter(this, void 0, void 0, function* () {
            const handler = this.getOperationHandler(operation);
            return handler.handle(operation);
        });
    }
}
exports.Metaplex = Metaplex;
//# sourceMappingURL=Metaplex.js.map