import NodeBundlr, { WebBundlr } from '@bundlr-network/client';
import { StorageDriver } from './StorageDriver';
import { KeypairIdentityDriver } from '../identity/KeypairIdentityDriver';
import { PlanUploadMetadataOperation } from "../../modules";
import { PlanUploadMetadataUsingBundlrOperationHandler } from './PlanUploadMetadataUsingBundlrOperationHandler';
import { SolAmount } from "../../shared";
export const bundlrStorage = (options = {}) => ({
    install(metaplex) {
        metaplex.setStorage(new BundlrStorageDriver(metaplex, options));
        metaplex.register(PlanUploadMetadataOperation, PlanUploadMetadataUsingBundlrOperationHandler);
    },
});
export class BundlrStorageDriver extends StorageDriver {
    constructor(metaplex, options = {}) {
        super(metaplex);
        this.bundlr = null;
        this.options = {
            providerUrl: metaplex.connection.rpcEndpoint,
            ...options,
        };
    }
    async getBalance() {
        const bundlr = await this.getBundlr();
        const balance = await bundlr.getLoadedBalance();
        return SolAmount.fromLamports(balance);
    }
    async getPrice(...files) {
        const price = await this.getMultipliedPrice(files);
        return SolAmount.fromLamports(price);
    }
    async upload(file) {
        const [uri] = await this.uploadAll([file]);
        return uri;
    }
    async uploadAll(files) {
        await this.fund(files);
        const promises = files.map((file) => this.uploadFile(file));
        // TODO: withdraw any money left in the balance?
        return Promise.all(promises);
    }
    async needsFunding(files) {
        const bundlr = await this.getBundlr();
        const balance = await bundlr.getLoadedBalance();
        const price = await this.getMultipliedPrice(files);
        return price.isGreaterThan(balance);
    }
    async fund(files, skipBalanceCheck = false) {
        const bundlr = await this.getBundlr();
        const price = await this.getMultipliedPrice(files);
        if (skipBalanceCheck) {
            await bundlr.fund(price);
            return;
        }
        const balance = await bundlr.getLoadedBalance();
        if (price.isGreaterThan(balance)) {
            await bundlr.fund(price.minus(balance));
        }
    }
    async getMultipliedPrice(files) {
        var _a;
        const bundlr = await this.getBundlr();
        const bytes = files.reduce((total, file) => total + file.getBytes(), 0);
        const price = await bundlr.getPrice(bytes);
        return price.multipliedBy((_a = this.options.priceMultiplier) !== null && _a !== void 0 ? _a : 1.5).decimalPlaces(0);
    }
    async uploadFile(file) {
        const bundlr = await this.getBundlr();
        const { status, data } = await bundlr.uploader.upload(file.toBuffer(), file.getTagsWithContentType());
        if (status >= 300) {
            // TODO: Custom errors.
            throw new Error(`Failed to upload asset. Got status: ${status}.`);
        }
        return `https://arweave.net/${data.id}`;
    }
    async withdrawAll() {
        // TODO: Implement when available on Bundlr.
        throw new Error('Method not implemented.');
    }
    async getBundlr() {
        var _a, _b;
        if (this.bundlr)
            return this.bundlr;
        const currency = 'solana';
        const address = (_b = (_a = this.options) === null || _a === void 0 ? void 0 : _a.address) !== null && _b !== void 0 ? _b : 'https://node1.bundlr.network';
        const options = {
            timeout: this.options.timeout,
            providerUrl: this.options.providerUrl,
        };
        const identity = this.metaplex.identity();
        const bundlr = identity instanceof KeypairIdentityDriver
            ? new NodeBundlr(address, currency, identity.keypair.secretKey, options)
            : new WebBundlr(address, currency, identity, options);
        try {
            // Check for valid bundlr node.
            await bundlr.utils.getBundlerAddress(currency);
        }
        catch (error) {
            // TODO: Custom errors.
            throw new Error(`Failed to connect to bundlr ${address}.`);
        }
        if (bundlr instanceof WebBundlr) {
            try {
                // Try to initiate bundlr.
                await bundlr.ready();
            }
            catch (error) {
                console.error(error);
            }
            if (!bundlr.address) {
                // TODO: Custom errors.
                throw new Error('Failed to initiate Bundlr.');
            }
        }
        this.bundlr = bundlr;
        return bundlr;
    }
}
//# sourceMappingURL=BundlrStorageDriver.js.map